/*  pdnmesh - a 2D finite element solver
    Copyright (C) 2001-2005 Sarod Yatawatta <sarod@users.sf.net>  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  $Id: input.c,v 1.20 2005/04/26 05:16:47 sarod Exp $
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* IO rutines */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "types.h"

poly_edge *edge_array;
int nedges;
boundary bound; /* boundaries */
MY_DOUBLE g_xoff,g_yoff; /* x,y offsets */
MY_DOUBLE g_xscale,g_yscale; /* x,y scales */
MY_DOUBLE g_xrange,g_yrange; /* x,y ranges, used for contour plot */


/* skips comment lines */
static int
skip_lines(FILE *fin)
{

		int c;
		do {
		if ( ( c = getc(fin) ) == EOF )
		return(-1);
	/* handle empty lines */
	if ( c == '\n' )
	continue; /* next line */
	if ( (c != '#') ) {
	ungetc(c,fin);
	return(0);
	} else { /* skip this line */
	do {
	if ( ( c = getc(fin) ) == EOF )
	return(-1);
	} while (  c != '\n') ;
	}
	} while( 1 );
}

/* skips rest of line */
static int
skip_restof_line(FILE *fin)
{
	int c;
	do {
	if ( ( c = getc(fin) ) == EOF )
	return(-1);
	} while (  c != '\n') ;
	return(1);
}

/* reads the next string (isalphanumeric() contiguous set of characters)
  separated by spaces, tabs or a newline. If the last character read is newline
  1 is returned, else 0 returned. */
/* buffer is automatically adjusted is length is not enough */
static int
read_next_string(char *buff, int *buff_len, FILE *infd) {
   int k,c,flag;
   k = 0;
   /* intialize buffer */
   buff[0]='\0';
   /* skip leading white space */
   do {
   c=fgetc(infd);
   if(c=='\n' || c==EOF) return 1;
   } while(c != EOF && isblank(c));
   if(c=='\n' || c==EOF) return 1;
   /* now we have read a non whitespace character */
   buff[k++]=c;
  if (k==*buff_len) {          
    /* now we have run out of buffer */
    *buff_len += 30;
    if ((buff = (char*)realloc((void*)buff,sizeof(char)*(size_t)(*buff_len)))==NULL) {
  fprintf(stderr,"%s: %d: No free memory\n",__FILE__,__LINE__);
  exit(1);
  }
  }
   flag=0;
   while ( ((c = fgetc(infd)) != EOF ) && k < *buff_len) {
      if ( c == '\n') {  flag=1; break; }
      if ( isblank(c) ) {  break; }/* not end of line */ 
      buff[k++] = c;
      if (k==*buff_len) {          
       /* now we have run out of buffer */
       *buff_len += 30;
       if((buff = (char*)realloc((void*)buff,sizeof(char)*(size_t)(*buff_len)))==NULL) {
    fprintf(stderr,"%s: %d: no free memory\n",__FILE__,__LINE__);
    exit(1);
       }
      }
  }
  /* now c == blank , \n or EOF */
  if (k==*buff_len-2) {          
    /* now we have run out of buffer */
    *buff_len += 3;
    if((buff = (char*)realloc((void*)buff,sizeof(char)*(size_t)(*buff_len)))==NULL) {
    fprintf(stderr,"%s: %d: No free memory\n",__FILE__,__LINE__);
    exit(1);
   }
  }

  /* add trailing ';' to buffer and '\0' */
  buff[k++]=';';
  buff[k++]='\0';
  return flag;
}


/* read input coordinate file */
int
read_input_file(point **input_points, const char *infile)
{
 /* read points from infile and store coords in x,y, arrays */
 FILE *infd;
 int i,j,count;
	MY_DOUBLE xmax,xmin,ymax,ymin;
	int npoint, input_degree_of_freedom=1;
 poly_edge e; 
int buff_len,c;
char *buff;
 exp_nodetype *pp=0; /* pointer to get root of parse tree from lex/yacc */

 infd =fopen( infile, "r" );
 if ( infd == NULL ) {
   fprintf(stderr, "%s: %d: could not open file %s\n", __FILE__,__LINE__,infile);
   exit(1);
 }
	skip_lines(infd);
	/* numer of points */
 if ( fscanf(infd, "%d" , &npoint ) == EOF ) {
    fprintf(stderr, "%s: %d: could not read file %s\n", __FILE__,__LINE__,infile);
    exit(1);
 }
	skip_restof_line(infd);
	skip_lines(infd);
#ifdef DEBUG
  printf("input: reading %d points\n",npoint);
#endif

 /* allocate memory for points */
 if ((*input_points=(point*)malloc((size_t)(npoint+3)*sizeof(point)))==0) {
   fprintf(stderr, "%s: %d: no free memory\n", __FILE__,__LINE__);
   exit(1);
 }

	xmax=-1000;xmin=1000;
	ymax=-1000;ymin=1000;

 for (i=3; i<npoint+3; i++ ) {
		/* allocate memory for z array */
	if (input_degree_of_freedom) {
  if (((*input_points)[i].z=(MY_DOUBLE*)malloc((size_t)(input_degree_of_freedom)*sizeof(MY_DOUBLE)))==0) {
   fprintf(stderr, "%s: %d: no free memory\n", __FILE__,__LINE__);
   exit(1);
 }
	}
	/* read point data according to degree of freedom */
	 if ( input_degree_of_freedom==0 ) { /* wave equation, do not need this */
    if ( fscanf(infd,"%d "MDF" "MDF"  %d",&count, &((*input_points)[i].x),&((*input_points)[i].y),&j) == EOF ) {
      fprintf(stderr, "%s: %d: could not read file %s\n", __FILE__,__LINE__,infile);
    }
   } else if ( input_degree_of_freedom==1 ) { /* potentials */
    if ( fscanf(infd,"%d "MDF" "MDF" "MDF" %d",&count, &((*input_points)[i].x),&((*input_points)[i].y),&((*input_points)[i].z[0]),&j) == EOF ) {
      fprintf(stderr, "%s: %d: could not read file %s\n", __FILE__,__LINE__,infile);
    }
  } else { /* input_degree_of_freedom==2, elasticity problems etc. */
    if ( fscanf(infd,"%d "MDF" "MDF" "MDF" "MDF" %d",&count, &((*input_points)[i].x),&((*input_points)[i].y),&((*input_points)[i].z[0]),&((*input_points)[i].z[1]), &j) == EOF ) {
      fprintf(stderr, "%s: %d: could not read file %s\n", __FILE__,__LINE__,infile);
    }
  }

#ifdef DEBUG
  printf("input: read point %d as %d ("MDF", "MDF"): input degree_of_freedom=%d ",i,count,(*input_points)[i].x, (*input_points)[i].y,input_degree_of_freedom);
	if ( input_degree_of_freedom==0 ) {
  printf("no z values\n");
  } else if (input_degree_of_freedom==1) {
   printf("z[0]="MDF", j=%d\n",(*input_points)[i].z[0],j);
	} else {
   printf("z[0]="MDF", z[1]="MDF",j=%d\n",(*input_points)[i].z[0], (*input_points)[i].z[1], j);
	}
#endif
 skip_restof_line(infd);
 skip_lines(infd);

 if ( xmin > (*input_points)[i].x ) {
								xmin=(*input_points)[i].x;
	}
	if ( xmax < (*input_points)[i].x ) {
								xmax=(*input_points)[i].x;
	}
 if ( ymin > (*input_points)[i].y ) {
								ymin=(*input_points)[i].y;
	}
	if ( ymax < (*input_points)[i].y ) {
								ymax=(*input_points)[i].y;
	}
/* point type */
	if (j==1) {
					/* fixed point */
					(*input_points)[i].val=FX;
	} else {
					/* variable point */
					(*input_points)[i].val=VAR;
	}
 }
 g_xoff=-(xmax+xmin)*0.5;
 g_yoff=-(ymax+ymin)*0.5;
 g_xrange=(xmax-xmin)*0.5;
 g_yrange=(ymax-ymin)*0.5;

	 /* now find the maximum value of coordinates */
  ymax = ABS(ymax);
		xmax = ABS(xmax);
		xmin = ABS(xmin);
		ymin = ABS(ymin);
		if ( xmax < ymax )
			xmax = ymax;
		if ( xmax < ymin )
			xmax = ymin;
		if ( xmax < xmin  )
			xmax = xmin;
  g_xscale=xmax;
  g_yscale=xmax;
		/* first 3 points (-3M,-3M),(3M,0),(0,3M) */
		 for (i=3;i<npoint+3;i++) {
          (*input_points)[i].x+=(g_xoff);
          (*input_points)[i].y+=(g_yoff);

          (*input_points)[i].x/=(xmax);
          (*input_points)[i].y/=(xmax);
		 }
			(*input_points)[0].x=-3;(*input_points)[0].y=-3;
			(*input_points)[1].x=3;(*input_points)[1].y=0;
			(*input_points)[2].x=0;(*input_points)[2].y=3;


	/* now the edges */
	/* first number of edges */
   if ( fscanf(infd, "%d" , &nedges) == EOF ) {
    fprintf(stderr, "%s: %d: could not read file %s\n", __FILE__,__LINE__,infile);
    exit(1);
 }
 skip_restof_line(infd);
	skip_lines(infd);


#ifdef DEBUG
  printf("input: reading %d edges\n",nedges);
#endif

 /* allocate memory for edge array */
 if ((edge_array =(poly_edge *)malloc((size_t)nedges*sizeof(poly_edge)))==0){
   fprintf(stderr, "%s: %d: no free memory\n", __FILE__,__LINE__);
   exit(1);
 }

 for (i=0; i< nedges; i++) {
 /* FIXME - edges need not be in number order */
  if ( fscanf(infd, "%d %d %d %d" , &count, &(edge_array[i].p1), &(edge_array[i].p2), &j) == EOF ) {
	      fprintf(stderr, "%s: %d: could not read file %s\n", __FILE__,__LINE__,infile);
				exit(1);
		}
 skip_restof_line(infd);
	skip_lines(infd);


		/* account for the 3 outer points, so increment point numbers */
   edge_array[i].p1+=3;edge_array[i].p2+=3;
			/* dirichlet or neumann ? */
			if (j==1) {
				edge_array[i].type=DR;
			} else {
				edge_array[i].type=NU;
			}
#ifdef DEBUG
  printf("input: edge %d (%d,%d) ?%d\n",i,edge_array[i].p1,edge_array[i].p2,j);
#endif

 }

	/* read the boundaries */
 /* first number of boundaries */
   if ( fscanf(infd, "%d" , &bound.npoly) == EOF ) {
    fprintf(stderr, "%s: %d: could not read file %s\n", __FILE__,__LINE__,infile);
    exit(1);
 }
 skip_restof_line(infd);
	skip_lines(infd);


	/* allocate memory for boundaries */
 if ((bound.poly_array=(polygon*)malloc((size_t)bound.npoly*sizeof(polygon)))==0){
   fprintf(stderr, "%s: %d: no free memory\n", __FILE__,__LINE__);
   exit(1);
 }

/* setup buffer to read expressions */
 buff_len = 32;
 if((buff = (char*)malloc(sizeof(char)*(size_t)(buff_len)))==NULL) {
        fprintf(stderr,"%s: %d: No free memory\n",__FILE__,__LINE__);
        exit(1);
 }

for (count=0; count< bound.npoly; count++) {
  skip_lines(infd);
  /* initialize polygon */
	init_poly(&bound.poly_array[count]);

	/* now read the outer boundary */
 /* boundary format :
   number of edges, hollow?, mu, epsilon, rho 
    OR
   number of edge, hollow?, 1/mu or epsilon, rho - old format */
	/* first, number of edges, hollow?*/
	if ( fscanf(infd, "%d %d", &i, 
							&bound.poly_array[count].hollow) == EOF ) {
    fprintf(stderr, "%s: %d: could not read file %s\n", __FILE__,__LINE__,infile);
    exit(1);
 }

 /* next read mu */
 memset(buff,0,buff_len);
 bound.poly_array[count].muexp=0;
 c=read_next_string(buff,&buff_len,infd);
#ifdef DEBUG
 printf("%d, mu=%s\n",count, buff);
#endif
 /* if c==1, '\n' was detected and we cannot read any more */
 if(c != 1) {
  /* parse buffer and extract mu */
  /* mu is an expression */
   pp=0;
   bound.poly_array[count].mu=1.0;
	 equation_parse(buff,&pp);
	 bound.poly_array[count].muexp= pp;
   if((bound.poly_array[count].mustr
       =(char*)malloc(sizeof(char)*(size_t)(strlen(buff)+1)))==NULL) {
       fprintf(stderr,"%s: %d: No free memory\n",__FILE__,__LINE__);
       exit(1);
   }
   strcpy(bound.poly_array[count].mustr, buff);
 /* if we are reading 1/mu get the reciprocal (mu cannot be below 1) */

 /* next read epsilon or rho. we do not know this at this point,
  but we assume it is epsilon */
 memset(buff,0,buff_len);
 bound.poly_array[count].epsexp=0;
 c=read_next_string(buff,&buff_len,infd);
#ifdef DEBUG
 printf("%d, eps=%s\n",count, buff);
#endif
  /* parse buffer and extract mu */
  /* mu is an expression */
   pp=0;
   bound.poly_array[count].epsilon=1.0;
	 equation_parse(buff,&pp);
	 bound.poly_array[count].epsexp= pp;
   if((bound.poly_array[count].epsstr=(char*)malloc(sizeof(char)*(size_t)(strlen(buff)+1)))==NULL) {
     fprintf(stderr,"%s: %d: No free memory\n",__FILE__,__LINE__);
     exit(1);
   }
   strcpy(bound.poly_array[count].epsstr, buff);
   /* if c==1, '\n' was detected and we cannot read any more */
   if(c != 1) {

  /* now read the expression for rho */
    memset(buff,0,buff_len);
    bound.poly_array[count].rhoexp=0;
    c=read_next_string(buff,&buff_len,infd);
#ifdef DEBUG
 printf("%d, rho=%s\n",count,buff);
#endif
        pp=0;
        bound.poly_array[count].rho=0;
				equation_parse(buff,&pp);
				bound.poly_array[count].rhoexp= pp;
       if((bound.poly_array[count].rhostr=(char*)malloc(sizeof(char)*(size_t)(strlen(buff)+1)))==NULL) {
         fprintf(stderr,"%s: %d: No free memory\n",__FILE__,__LINE__);
        exit(1);
       }
       strcpy(bound.poly_array[count].rhostr, buff);
 } else {
   /* c==1 means end of line. so we have read rho, NOT epsilon */
   /* copy back everything to rho */
   bound.poly_array[count].rho=bound.poly_array[count].epsilon;
   bound.poly_array[count].rhoexp=bound.poly_array[count].epsexp;
   bound.poly_array[count].rhostr=bound.poly_array[count].epsstr;

   bound.poly_array[count].epsilon=1.0;
   bound.poly_array[count].epsexp=0;
   /*bound.poly_array[count].epsstr="1.0;"; */
   if((bound.poly_array[count].epsstr=(char*)malloc(sizeof(char)*(size_t)(5)))==NULL) {
     fprintf(stderr,"%s: %d: No free memory\n",__FILE__,__LINE__);
     exit(1);
   }
   strcpy(bound.poly_array[count].epsstr, "1.0;");


 }
 } else {
    fprintf(stderr,"%s: %d: invalid numeric value for mu %s\n",__FILE__,__LINE__,buff);
   bound.poly_array[count].mu=1.0;
   bound.poly_array[count].epsilon=1.0;
   bound.poly_array[count].rho=0.0;

   bound.poly_array[count].muexp=0;
   bound.poly_array[count].epsexp=0;
   bound.poly_array[count].rhoexp=0;

   /*bound.poly_array[count].mustr="1.0;"; */
   if((bound.poly_array[count].mustr=(char*)malloc(sizeof(char)*(size_t)(5)))==NULL) {
     fprintf(stderr,"%s: %d: No free memory\n",__FILE__,__LINE__);
     exit(1);
   }
   strcpy(bound.poly_array[count].mustr, "1.0;");
   /*bound.poly_array[count].epsstr="1.0;"; */
   if((bound.poly_array[count].epsstr=(char*)malloc(sizeof(char)*(size_t)(5)))==NULL) {
     fprintf(stderr,"%s: %d: No free memory\n",__FILE__,__LINE__);
     exit(1);
   }
   strcpy(bound.poly_array[count].epsstr, "1.0;");
   /*bound.poly_array[count].rhostr="1.0;"; */
   if((bound.poly_array[count].rhostr=(char*)malloc(sizeof(char)*(size_t)(5)))==NULL) {
     fprintf(stderr,"%s: %d: No free memory\n",__FILE__,__LINE__);
     exit(1);
   }
   strcpy(bound.poly_array[count].rhostr, "1.0;");
 }
#ifdef DEBUG
 printf("boundary %d, mu=%s, epsilon=%s, rho=%s\n",count, bound.poly_array[count].mustr,
  bound.poly_array[count].epsstr,bound.poly_array[count].rhostr);
#endif

 skip_lines(infd);

#ifdef DEBUG
   printf("input: boundary %d has %d edges, hollow=%d, 1/mu="MDF", rho="MDF"\n",
																			count,i,bound.poly_array[count].hollow,
																			bound.poly_array[count].mu,bound.poly_array[count].rho);
#endif
/* read edge numbers per line */
	for (; i; i--){
    if ( fscanf(infd,"%d",&j) == EOF ) {
      fprintf(stderr, "%s: %d: could not read file %s\n", __FILE__,__LINE__,infile);
			exit(1);
    }
	e.p1 =edge_array[j].p1; e.p2=edge_array[j].p2; 
	e.type=edge_array[j].type;
	insert_edge_to_poly(&bound.poly_array[count],&e);
#ifdef DEBUG
   printf("input: %d of %d is edge %d :(%d  %d)\n",i,bound.poly_array[count].nedges,j,e.p1,e.p2);
#endif

	}
	/* arrange polygon edges */
#ifdef DEBUG
	printf("input: rearranging polygon %d\n",count);
#endif
 arrange_polgon_edges(&bound.poly_array[count]);
}

	 free(buff); 
   fclose(infd);
   return(npoint+3);
}


/* function to get the expression for rho from input and build parse tree */
/* here is how its called: 1: equation_parse(buffer)->call yyparse() with buffer
 * as input->parses buffer and builds tree->calls insert_expression to 
 * attach tree to current boundary globally */
/*int
insert_expression(exp_nodetype *p, MY_INT boundary_no)
{
				      bound.poly_array[boundary_no].rhoexp = p;
							return(0);
}*/
